#!/usr/bin/perl
# by Claudio Fanelli
# probably it has a lot of bugs ;-)

# based on a work of Graham Ellis ( you can find it at
# http://www.wellho.net/forum/Perl-Programming/HTTP.html )

# LIMITS:
# supports only GET e POST
# all the web pages must be put in /dev/shm/, except the cgi, which need to 
# be put in the same directory of perl_webserver
# it does not fork (one request at a time)

# GLOBAL VARS (for use in CGI):
# %_POST : contains post variables
# %_GET : contains get variables
# %_REQUEST : contains get&post variables
# %_COOKIES : contains cookies variables
# %sent_headers : contains headers sent by the client browser (the names are 
#                 lowercase)
# %sent_lines : contains all the lines sent by the client browser
# client_ipaddr: ip address of the client

# exported functions:
#   HTMLencode()  #encodes a LIST of strings in url format (<space>=%20)

use warnings;
use Socket;
use POSIX ":sys_wait_h";

# read the configuration file
chdir "/usr/local/perl_webserver";
require("config_pw.pl");
foreach my $c (@cgi) {
    require($c);
}
$exec_dir=`pwd`;
chop $exec_dir;
chdir "/dev/shm/";

sub HTMLencode {
    $cont=0;    
    foreach (@_) {
	$stringa=$_;
	$stringa =~ s/([^a-zA-Z0-1])/sprintf("%%%2x",ord($1))/eg;
	$ris[$cont++]=$stringa;
    }
    if(wantarray) {
      return @ris;
    } else {
	return join "",@ris;
    }
}

sub get_vars {
    # this subroutine splits $data in chuncks separated by $separator, 
    # putting that data in $ref_hash (an hash reference)
    # it is used for obtaining GET vars, POST vars and cookies
    $data=shift || return ;
    $ref_hash=shift || return;
    $separator= shift || "&";
    @pezzi = split(/$separator/,$data);
    foreach my $pezzo (@pezzi) {
	($f,$v) = split(/=/,$pezzo,2);
	# trasformiamo i caratteri %hex in caratteri stampabili
	while( $v =~ /%([0-9a-fA-F]{1,2})/ ) {
	    $ris=chr( hex($1) );
	    $v =~ s/%$1/$ris/;
	    pos($v)=0; # per sicurezza facciamo ripartire il match dalla 
	    # posizione 0, visto che abbiamo cambiato la stringa, 
	    # probabilmente e' inutile
	}	
	$v =~ s/\+/ /; # trasformiamo i + in spazi
	$ref_hash->{lc($f)} = $v;
    }
}

# Set up listener
$proto = getprotobyname("tcp");
socket (Server, PF_INET, SOCK_STREAM, $proto) || 
  die ("cannot  create socket");
setsockopt(Server, SOL_SOCKET,SO_REUSEADDR,1) || 
  die ("cannot set setsock options");
$pbind = sockaddr_in($on_port,INADDR_ANY) ||  
  die ("cannot set socket address");
bind(Server,$pbind) || die ("cannot bind");
listen(Server,SOMAXCONN) || die ("cannot listen");
    
# otteniamo i mime types
open(IN,"</etc/mime.types");
while(<IN>) {
    if (/^#/) {next;}
    ($mime,$exts)=split(/\s+/,$_,2);
    if($exts ne "") {
	@estensioni=split(/\s/,$exts);
	foreach my $estensione (@estensioni) {
	    $mime_types{$estensione}=$mime;
	}
    }
}
close IN;
    
@response = <DATA>;
my $out=join "",@response;
# await a contact
while ($client_ipaddr = accept(Client, Server)) {
    if( fork== 0) { # figlio
	while (<Client>) { # leggiamo l'header http inviato, e mettiamo in 
	    # %sent_headers
	    push @sent_lines,$_;
	    if ($_ =~ /^\s*$/) { 
		last; # l'header termina con una linea vuota
	    }
	    if(/:\s+/) { # poniamo l'elemento attuale in %sent_headers, se non
		# e' la linea col GET o POST
		($f,$v) = split(/:\s+/,$_,2);
		$v=~ s/\n|\r//g;
		$sent_headers{lc($f)} = $v;
	    }
	       
	}
	# guardiamo se ci sono cookie, in caso affermativo settiamo la
	# variabile %_COOKIE
	%_COOKIE=();
	if($sent_headers{"cookie"} ne "" ) {
	    get_vars($sent_headers{"cookie"},\%_COOKIE,";.");
	}
	    
	# se il methodo e' POST, otteniamo il nome del file e le variabili di POST
	%_POST=();
	if ($sent_lines[0] =~ /^POST \/([^? ]+)/) {
	    $file=$1;
	    read(Client,$data,$sent_headers{"content-length"});
	    get_vars($data,\%_POST);
	} 

	# se il methodo e' GET, otteniamo il nome del file e le variabili di GET
	%_GET=();
	if ($sent_lines[0] =~ /^GET \/([^? ]+)(\?([^ ]+))?/) {
	    $get=1;
	    $file=$1;
	    if($3 ne "") {
		get_vars($3,\%_GET);
	    }
	}
	#creiamo REQUEST fondendo i due array. GET prima di post
	%_REQUEST=();
	foreach (keys %_GET) {
	    $_REQUEST{$_}=$_GET{$_};
	}
	foreach (keys %_POST) {
	    $_REQUEST{$_}=$_POST{$_};
	}

	$file=~ s/\.\.\///g; # eliminiamo tutti i ../, per ridurre problemi di
	#  sicurezza ora stabiliamo  se e' una cgi registrata, accessibile sia
	# in post che in get, o una pagina statica, accessibile solo in GET
	if( $file =~/(.*)\.pl$/ && grep(/$file/,@cgi)) {  # stabiliamo se e' 
	    # una cgi registrata
	    $out.=&$1();
	} elsif( $get && -s $file) { # stabiliamo se e' un file da prelevare
	    # in GET
	    $file =~ /\.([a-z0-9]{3,4})$/i;
	    $ext=$1;
	    $mime=$mime_types{$ext};
	    if($mime eq "") {
		$mime="application/octet-stream";
	    }
	    $out.="Content-type: ".$mime."\n";	    
	    if(!open(IN,"<$file") ) {		
		$out="HTTP/1.0 404 Not Found\n\n<html><body>".
		  "<p>pagina sconosciuta:$file</p></body></html>\n";
	    } else {
		$len= -s $file;
		undef $/;
		$letto=<IN>;
		close(IN);
		$out.="Content-length: ".$len."\n\n";
		$out.=$letto;
	    }
	} else { # non e' ne' una pagina cgi, ne' un file da prendere in GET
	    $out="HTTP/1.0 404 Not Found\n\npagina sconosciuta:$file\n";
	}
	print Client $out;
	close Client;
	$#sent_lines = -1;
	undef %sent_headers;
	undef %_POST;
	undef %_GET;
	exit(0);
    }
    # liberiamoci degli zombie
    do {
	$kid = waitpid(-1, WNOHANG);
    } until $kid > 0;
    close Client;
}
__END__
HTTP/1.0 200 OK
Date: Sun, 18 Apr 1999 13:59:18 GMT
Server: vulcano 1.0
Expires: Fri, 30 Oct 1998 14:19:41 GMT
